@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * the actual codes (C syntax, i.e. octal):
@@ -33,11 +33,11 @@
  * \1ab          - a ModRM, calculated on EA in operand a, with the spare
  *                 field the register value of operand b.
  * \140..\143    - an immediate word or signed byte for operand 0..3
- * \144..\147    - or 2 (s-field) into next opcode byte if operand 0..3
- *		    is a signed byte rather than a word.
+ * \144..\147    - or 2 (s-field) into opcode byte if operand 0..3
+ *		    is a signed byte rather than a word.  Opcode byte follows.
  * \150..\153     - an immediate dword or signed byte for operand 0..3
- * \154..\157     - or 2 (s-field) into next opcode byte if operand 0..3
- *		    is a signed byte rather than a dword.
+ * \154..\157    - or 2 (s-field) into opcode byte if operand 0..3
+ *		    is a signed byte rather than a word.  Opcode byte follows.
  * \160..\163    - this instruction uses DREX rather than REX, with the
  *		   OC0 field set to 0, and the dest field taken from
  *                 operand 0..3.
@@ -54,6 +54,10 @@
  * \311          - indicates fixed 32-bit address size, i.e. optional 0x67.
  * \312          - (disassembler only) marker on LOOP, LOOPxx instructions.
  * \313          - indicates fixed 64-bit address size, 0x67 invalid.
+ * \314          - (disassembler only) invalid with REX.B
+ * \315          - (disassembler only) invalid with REX.X
+ * \316          - (disassembler only) invalid with REX.R
+ * \317          - (disassembler only) invalid with REX.W
  * \320          - indicates fixed 16-bit operand size, i.e. optional 0x66.
  * \321          - indicates fixed 32-bit operand size, i.e. optional 0x66.
  * \322          - indicates that this instruction is only valid when the
@@ -107,8 +111,8 @@ static efunc errfunc;
 static struct ofmt *outfmt;
 static ListGen *list;
 
-static int32_t calcsize(int32_t, int32_t, int, insn *, const char *);
-static void gencode(int32_t, int32_t, int, insn *, const char *, int32_t);
+static int64_t calcsize(int32_t, int64_t, int, insn *, const char *);
+static void gencode(int32_t, int64_t, int, insn *, const char *, int64_t);
 static int matches(const struct itemplate *, insn *, int bits);
 static int32_t regflag(const operand *);
 static int32_t regval(const operand *);
@@ -152,10 +156,10 @@ static const char *size_name(int size)
 static void warn_overflow(int size, int64_t data)
 {
     if (size < 8) {
-	int64_t lim = (1 << (size*8))-1;
+	int64_t lim = ((int64_t)1 << (size*8))-1;
 
 	if (data < ~lim || data > lim)
-	    errfunc(ERR_WARNING, "%s data exceeds bounds", size_name(size));
+	    errfunc(ERR_WARNING | ERR_WARN_NOV, "%s data exceeds bounds", size_name(size));
     }
 }
 /*
@@ -163,45 +167,33 @@ static void warn_overflow(int size, int64_t data)
  * in order to pass a copy of the data off to the listing file
  * generator at the same time.
  */
-static void out(int32_t offset, int32_t segto, const void *data,
-                uint32_t type, int32_t segment, int32_t wrt)
+static void out(int64_t offset, int32_t segto, const void *data,
+                enum out_type type, uint64_t size,
+		int32_t segment, int32_t wrt)
 {
     static int32_t lineno = 0;     /* static!!! */
     static char *lnfname = NULL;
+    uint8_t p[8];
 
-    if ((type & OUT_TYPMASK) == OUT_ADDRESS) {
-        if (segment != NO_SEG || wrt != NO_SEG) {
-            /*
-             * This address is relocated. We must write it as
-             * OUT_ADDRESS, so there's no work to be done here.
-             */
-            list->output(offset, data, type);
-        } else {
-            uint8_t p[8], *q = p;
-            /*
-             * This is a non-relocated address, and we're going to
-             * convert it into RAWDATA format.
-             */
-            if ((type & OUT_SIZMASK) == 4) {
-                WRITELONG(q, *(int32_t *)data);
-                list->output(offset, p, OUT_RAWDATA + 4);
-            } else if ((type & OUT_SIZMASK) == 8) {
-                WRITEDLONG(q, *(int64_t *)data);
-                list->output(offset, p, OUT_RAWDATA + 8);
-            } else {
-                WRITESHORT(q, *(int32_t *)data);
-                list->output(offset, p, OUT_RAWDATA + 2);
-            }
-        }
-    } else if ((type & OUT_TYPMASK) == OUT_RAWDATA) {
-        list->output(offset, data, type);
-    } else if ((type & OUT_TYPMASK) == OUT_RESERVE) {
-        list->output(offset, NULL, type);
-    } else if ((type & OUT_TYPMASK) == OUT_REL2ADR ||
-               (type & OUT_TYPMASK) == OUT_REL4ADR) {
-        list->output(offset, data, type);
+    if (type == OUT_ADDRESS && segment == NO_SEG && wrt == NO_SEG) {
+	/*
+	 * This is a non-relocated address, and we're going to
+	 * convert it into RAWDATA format.
+	 */
+	uint8_t *q = p;
+
+	if (size > 8) {
+	    errfunc(ERR_PANIC, "OUT_ADDRESS with size > 8");
+	    return;
+	}
+	
+	WRITEADDR(q, *(int64_t *)data, size);
+	data = p;
+	type = OUT_RAWDATA;
     }
 
+    list->output(offset, data, type, size);
+
     /*
      * this call to src_get determines when we call the
      * debug-format-specific "linenum" function
@@ -215,13 +207,13 @@ static void out(int32_t offset, int32_t segto, const void *data,
         outfmt->current_dfmt->linenum(lnfname, lineno, segto);
     }
 
-    outfmt->output(segto, data, type, segment, wrt);
+    outfmt->output(segto, data, type, size, segment, wrt);
 }
 
-static int jmp_match(int32_t segment, int32_t offset, int bits,
+static int jmp_match(int32_t segment, int64_t offset, int bits,
                      insn * ins, const char *code)
 {
-    int32_t isize;
+    int64_t isize;
     uint8_t c = code[0];
 
     if (c != 0370 && c != 0371)
@@ -243,17 +235,17 @@ static int jmp_match(int32_t segment, int32_t offset, int bits,
     return 0;
 }
 
-int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
+int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
               insn * instruction, struct ofmt *output, efunc error,
               ListGen * listgen)
 {
     const struct itemplate *temp;
     int j;
     int size_prob;
-    int32_t insn_end;
+    int64_t insn_end;
     int32_t itimes;
-    int32_t start = offset;
-    int32_t wsize = 0;             /* size for DB etc. */
+    int64_t start = offset;
+    int64_t wsize = 0;             /* size for DB etc. */
 
     errfunc = error;            /* to pass to other functions */
     cpu = cp;
@@ -302,26 +294,27 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                         else {
                             uint8_t out_byte = e->offset;
                             out(offset, segment, &out_byte,
-                                OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                                OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                         }
                     } else if (wsize > 8) {
                         errfunc(ERR_NONFATAL, "integer supplied to a DT or DO"
                                 " instruction");
                     } else
                         out(offset, segment, &e->offset,
-                            OUT_ADDRESS + wsize, e->segment, e->wrt);
+                            OUT_ADDRESS, wsize, e->segment, e->wrt);
                     offset += wsize;
                 } else if (e->type == EOT_DB_STRING) {
                     int align;
 
                     out(offset, segment, e->stringval,
-                        OUT_RAWDATA + e->stringlen, NO_SEG, NO_SEG);
+                        OUT_RAWDATA, e->stringlen, NO_SEG, NO_SEG);
                     align = e->stringlen % wsize;
 
                     if (align) {
                         align = wsize - align;
-                        out(offset, segment, "\0\0\0\0\0\0\0\0",
-                            OUT_RAWDATA + align, NO_SEG, NO_SEG);
+                        out(offset, segment,
+			    "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0",
+                            OUT_RAWDATA, align, NO_SEG, NO_SEG);
                     }
                     offset += e->stringlen + align;
                 }
@@ -331,7 +324,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                  * Dummy call to list->output to give the offset to the
                  * listing module.
                  */
-                list->output(offset, NULL, OUT_RAWDATA);
+                list->output(offset, NULL, OUT_RAWDATA, 0);
                 list->uplevel(LIST_TIMES);
             }
         }
@@ -393,7 +386,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
              * Dummy call to list->output to give the offset to the
              * listing module.
              */
-            list->output(offset, NULL, OUT_RAWDATA);
+            list->output(offset, NULL, OUT_RAWDATA, 0);
             list->uplevel(LIST_INCBIN);
             while (t--) {
                 int32_t l;
@@ -416,7 +409,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                         t = 0;  /* Try to exit cleanly */
                         break;
                     }
-                    out(offset, segment, buf, OUT_RAWDATA + m,
+                    out(offset, segment, buf, OUT_RAWDATA, m,
                         NO_SEG, NO_SEG);
                     l -= m;
                 }
@@ -427,7 +420,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                  * Dummy call to list->output to give the offset to the
                  * listing module.
                  */
-                list->output(offset, NULL, OUT_RAWDATA);
+                list->output(offset, NULL, OUT_RAWDATA, 0);
                 list->uplevel(LIST_TIMES);
                 list->downlevel(LIST_TIMES);
             }
@@ -450,7 +443,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
 
         if (m == 100) {         /* matches! */
             const char *codes = temp->code;
-            int32_t insn_size = calcsize(segment, offset, bits,
+            int64_t insn_size = calcsize(segment, offset, bits,
                                       instruction, codes);
             itimes = instruction->times;
             if (insn_size < 0)  /* shouldn't be, on pass two */
@@ -475,21 +468,21 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                         case R_CS:
                             if (bits == 64) {
                                 error(ERR_WARNING,
-                                      "cs segment base ignored in 64-bit mode");
+                                      "cs segment base generated, but will be ignored in 64-bit mode");
                             }
                             c = 0x2E;
                             break;
                         case R_DS:
                             if (bits == 64) {
                                 error(ERR_WARNING,
-                                      "ds segment base ignored in 64-bit mode");
+                                      "ds segment base generated, but will be ignored in 64-bit mode");
                             }
                             c = 0x3E;
                             break;
                         case R_ES:
                            if (bits == 64) {
                                 error(ERR_WARNING,
-                                      "es segment base ignored in 64-bit mode");
+                                      "es segment base generated, but will be ignored in 64-bit mode");
                            }
                             c = 0x26;
                             break;
@@ -502,7 +495,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                         case R_SS:
                             if (bits == 64) {
                                 error(ERR_WARNING,
-                                      "ss segment base ignored in 64-bit mode");
+                                      "ss segment base generated, but will be ignored in 64-bit mode");
                             }
                             c = 0x36;
                             break;
@@ -553,7 +546,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                             error(ERR_PANIC, "invalid instruction prefix");
                         }
                         if (c != 0) {
-                            out(offset, segment, &c, OUT_RAWDATA + 1,
+                            out(offset, segment, &c, OUT_RAWDATA, 1,
                                 NO_SEG, NO_SEG);
                             offset++;
                         }
@@ -567,7 +560,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
                          * Dummy call to list->output to give the offset to the
                          * listing module.
                          */
-                        list->output(offset, NULL, OUT_RAWDATA);
+                        list->output(offset, NULL, OUT_RAWDATA, 0);
                         list->uplevel(LIST_TIMES);
                     }
                 }
@@ -603,7 +596,7 @@ int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
     return 0;
 }
 
-int32_t insn_size(int32_t segment, int32_t offset, int bits, uint32_t cp,
+int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
                insn * instruction, efunc error)
 {
     const struct itemplate *temp;
@@ -723,7 +716,7 @@ int32_t insn_size(int32_t segment, int32_t offset, int bits, uint32_t cp,
 
         if (m == 100) {
             /* we've matched an instruction. */
-            int32_t isize;
+            int64_t isize;
             const char *codes = temp->code;
             int j;
 
@@ -782,10 +775,10 @@ static int is_sbyte(insn * ins, int op, int size)
     return ret && v >= -128L && v <= 127L;
 }
 
-static int32_t calcsize(int32_t segment, int32_t offset, int bits,
+static int64_t calcsize(int32_t segment, int64_t offset, int bits,
                      insn * ins, const char *codes)
 {
-    int32_t length = 0;
+    int64_t length = 0;
     uint8_t c;
     int rex_mask = ~0;
     struct operand *opx;
@@ -915,7 +908,7 @@ static int32_t calcsize(int32_t segment, int32_t offset, int bits,
         case 0145:
         case 0146:
         case 0147:
-            codes += 2;
+            codes++;
             length++;
             break;
         case 0150:
@@ -928,7 +921,7 @@ static int32_t calcsize(int32_t segment, int32_t offset, int bits,
         case 0155:
         case 0156:
         case 0157:
-            codes += 2;
+            codes++;
             length++;
             break;
 	case 0160:
@@ -972,6 +965,11 @@ static int32_t calcsize(int32_t segment, int32_t offset, int bits,
 		has_prefix(ins, PPS_ASIZE, P_A32))
 		return -1;
             break;
+	case 0314:
+	case 0315:
+	case 0316:
+	case 0317:
+	    break;
         case 0320:
             length += (bits != 16);
             break;
@@ -1001,13 +999,11 @@ static int32_t calcsize(int32_t segment, int32_t offset, int bits,
         case 0335:
 	    break;
         case 0340:
-        case 0341:
-        case 0342:
             if (ins->oprs[0].segment != NO_SEG)
                 errfunc(ERR_NONFATAL, "attempt to reserve non-constant"
                         " quantity of BSS space");
             else
-                length += ins->oprs[0].offset << (c & 3);
+                length += ins->oprs[0].offset;
             break;
 	case 0364:
 	case 0365:
@@ -1092,13 +1088,13 @@ static int32_t calcsize(int32_t segment, int32_t offset, int bits,
 #define EMIT_REX()							\
     if (!(ins->rex & REX_D) && (ins->rex & REX_REAL) && (bits == 64)) {	\
 	ins->rex = (ins->rex & REX_REAL)|REX_P;				\
-	out(offset, segment, &ins->rex, OUT_RAWDATA+1, NO_SEG, NO_SEG); \
+	out(offset, segment, &ins->rex, OUT_RAWDATA, 1, NO_SEG, NO_SEG); \
 	ins->rex = 0;							\
 	offset += 1; \
     }
 
-static void gencode(int32_t segment, int32_t offset, int bits,
-                    insn * ins, const char *codes, int32_t insn_end)
+static void gencode(int32_t segment, int64_t offset, int bits,
+                    insn * ins, const char *codes, int64_t insn_end)
 {
     static char condval[] = {   /* conditional opcodes */
         0x7, 0x3, 0x2, 0x6, 0x2, 0x4, 0xF, 0xD, 0xC, 0xE, 0x6, 0x2,
@@ -1107,7 +1103,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
     };
     uint8_t c;
     uint8_t bytes[4];
-    int32_t size;
+    int64_t size;
     int64_t data;
     struct operand *opx;
 
@@ -1119,7 +1115,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 02:
         case 03:
 	    EMIT_REX();
-            out(offset, segment, codes, OUT_RAWDATA + c, NO_SEG, NO_SEG);
+            out(offset, segment, codes, OUT_RAWDATA, c, NO_SEG, NO_SEG);
             codes += c;
             offset += c;
             break;
@@ -1143,7 +1139,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                 errfunc(ERR_PANIC,
                         "bizarre 8086 segment register received");
             }
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
             break;
 
@@ -1160,7 +1156,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                 errfunc(ERR_PANIC,
                         "bizarre 386 segment register received");
             }
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
             break;
 
@@ -1170,7 +1166,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 	case 013:
 	    EMIT_REX();
             bytes[0] = *codes++ + ((regval(opx)) & 7);
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1179,16 +1175,17 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 016:
 	case 017:
             if (opx->offset < -128 || opx->offset > 127) {
-                errfunc(ERR_WARNING, "signed byte value exceeds bounds");
+                errfunc(ERR_WARNING | ERR_WARN_NOV,
+			"signed byte value exceeds bounds");
             }
 
             if (opx->segment != NO_SEG) {
                 data = opx->offset;
-                out(offset, segment, &data, OUT_ADDRESS + 1,
+                out(offset, segment, &data, OUT_ADDRESS, 1,
                     opx->segment, opx->wrt);
             } else {
                 bytes[0] = opx->offset;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG,
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
             }
             offset += 1;
@@ -1199,15 +1196,16 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 022:
 	case 023:
             if (opx->offset < -256 || opx->offset > 255) {
-                errfunc(ERR_WARNING, "byte value exceeds bounds");
+                errfunc(ERR_WARNING | ERR_WARN_NOV,
+			"byte value exceeds bounds");
             }
             if (opx->segment != NO_SEG) {
                 data = opx->offset;
-                out(offset, segment, &data, OUT_ADDRESS + 1,
+                out(offset, segment, &data, OUT_ADDRESS, 1,
                     opx->segment, opx->wrt);
             } else {
                 bytes[0] = opx->offset;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG,
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
             }
             offset += 1;
@@ -1218,14 +1216,15 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 026:
 	case 027:
             if (opx->offset < 0 || opx->offset > 255)
-                errfunc(ERR_WARNING, "unsigned byte value exceeds bounds");
+                errfunc(ERR_WARNING | ERR_WARN_NOV,
+			"unsigned byte value exceeds bounds");
             if (opx->segment != NO_SEG) {
                 data = opx->offset;
-                out(offset, segment, &data, OUT_ADDRESS + 1,
+                out(offset, segment, &data, OUT_ADDRESS, 1,
                     opx->segment, opx->wrt);
             } else {
                 bytes[0] = opx->offset;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG,
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
             }
             offset += 1;
@@ -1238,7 +1237,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             data = opx->offset;
             if (opx->segment == NO_SEG && opx->wrt == NO_SEG)
 		warn_overflow(2, data);
-            out(offset, segment, &data, OUT_ADDRESS + 2,
+            out(offset, segment, &data, OUT_ADDRESS, 2,
                 opx->segment, opx->wrt);
             offset += 2;
             break;
@@ -1254,7 +1253,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             data = opx->offset;
             if (opx->segment == NO_SEG && opx->wrt == NO_SEG)
 		warn_overflow(size, data);
-            out(offset, segment, &data, OUT_ADDRESS + size,
+            out(offset, segment, &data, OUT_ADDRESS, size,
                 opx->segment, opx->wrt);
             offset += size;
             break;
@@ -1264,7 +1263,9 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 042:
 	case 043:
             data = opx->offset;
-            out(offset, segment, &data, OUT_ADDRESS + 4,
+            if (opx->segment == NO_SEG && opx->wrt == NO_SEG)
+		warn_overflow(4, data);
+            out(offset, segment, &data, OUT_ADDRESS, 4,
                 opx->segment, opx->wrt);
             offset += 4;
             break;
@@ -1278,7 +1279,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 	    if (opx->segment == NO_SEG &&
 		opx->wrt == NO_SEG)
 		warn_overflow(size, data);
-            out(offset, segment, &data, OUT_ADDRESS + size,
+            out(offset, segment, &data, OUT_ADDRESS, size,
                 opx->segment, opx->wrt);
             offset += size;
             break;
@@ -1294,7 +1295,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             if (data > 127 || data < -128)
                 errfunc(ERR_NONFATAL, "short jump is out of range");
             bytes[0] = data;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1303,7 +1304,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 056:
 	case 057:
             data = (int64_t)opx->offset;
-            out(offset, segment, &data, OUT_ADDRESS + 8,
+            out(offset, segment, &data, OUT_ADDRESS, 8,
                 opx->segment, opx->wrt);
             offset += 8;
             break;
@@ -1315,12 +1316,12 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             if (opx->segment != segment) {
                 data = opx->offset;
                 out(offset, segment, &data,
-                    OUT_REL2ADR + insn_end - offset,
+                    OUT_REL2ADR, insn_end - offset,
                     opx->segment, opx->wrt);
             } else {
                 data = opx->offset - insn_end;
                 out(offset, segment, &data,
-                    OUT_ADDRESS + 2, NO_SEG, NO_SEG);
+                    OUT_ADDRESS, 2, NO_SEG, NO_SEG);
             }
             offset += 2;
             break;
@@ -1334,14 +1335,14 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             else
                 size = (bits == 16) ? 2 : 4;
             if (opx->segment != segment) {
-                int32_t reltype = (size == 2 ? OUT_REL2ADR : OUT_REL4ADR);
                 data = opx->offset;
-                out(offset, segment, &data, reltype + insn_end - offset,
-                    opx->segment, opx->wrt);
+                out(offset, segment, &data,
+		    size == 2 ? OUT_REL2ADR : OUT_REL4ADR,
+		    insn_end - offset, opx->segment, opx->wrt);
             } else {
                 data = opx->offset - insn_end;
                 out(offset, segment, &data,
-                    OUT_ADDRESS + size, NO_SEG, NO_SEG);
+                    OUT_ADDRESS, size, NO_SEG, NO_SEG);
             }
             offset += size;
             break;
@@ -1353,12 +1354,12 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             if (opx->segment != segment) {
                 data = opx->offset;
                 out(offset, segment, &data,
-                    OUT_REL4ADR + insn_end - offset,
+                    OUT_REL4ADR, insn_end - offset,
                     opx->segment, opx->wrt);
             } else {
                 data = opx->offset - insn_end;
                 out(offset, segment, &data,
-                    OUT_ADDRESS + 4, NO_SEG, NO_SEG);
+                    OUT_ADDRESS, 4, NO_SEG, NO_SEG);
             }
             offset += 4;
             break;
@@ -1371,7 +1372,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                 errfunc(ERR_NONFATAL, "value referenced by FAR is not"
                         " relocatable");
             data = 0L;
-            out(offset, segment, &data, OUT_ADDRESS + 2,
+            out(offset, segment, &data, OUT_ADDRESS, 2,
                 outfmt->segbase(1 + opx->segment),
                 opx->wrt);
             offset += 2;
@@ -1384,14 +1385,14 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             data = opx->offset;
             if (is_sbyte(ins, c & 3, 16)) {
                 bytes[0] = data;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG,
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
                 offset++;
             } else {
                 if (opx->segment == NO_SEG &&
                     opx->wrt == NO_SEG)
 		    warn_overflow(2, data);
-                out(offset, segment, &data, OUT_ADDRESS + 2,
+                out(offset, segment, &data, OUT_ADDRESS, 2,
                     opx->segment, opx->wrt);
                 offset += 2;
             }
@@ -1402,11 +1403,10 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0146:
 	case 0147:
 	    EMIT_REX();
-            codes++;
             bytes[0] = *codes++;
             if (is_sbyte(ins, c & 3, 16))
                 bytes[0] |= 2;  /* s-bit */
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
             break;
 
@@ -1417,11 +1417,11 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             data = opx->offset;
             if (is_sbyte(ins, c & 3, 32)) {
                 bytes[0] = data;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG,
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
                 offset++;
             } else {
-                out(offset, segment, &data, OUT_ADDRESS + 4,
+                out(offset, segment, &data, OUT_ADDRESS, 4,
                     opx->segment, opx->wrt);
                 offset += 4;
             }
@@ -1432,11 +1432,10 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0156:
 	case 0157:
 	    EMIT_REX();
-            codes++;
             bytes[0] = *codes++;
             if (is_sbyte(ins, c & 3, 32))
                 bytes[0] |= 2;  /* s-bit */
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
             break;
 
@@ -1453,7 +1452,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0170:
 	    EMIT_REX();
             bytes[0] = 0;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1463,7 +1462,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 		(ins->rex & REX_OC ? 0x08 : 0) |
 		(ins->rex & (REX_R|REX_X|REX_B));
 	    ins->rex = 0;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
 	    offset++;
 	    break;
 
@@ -1476,8 +1475,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0310:
             if (bits == 32 && !has_prefix(ins, PPS_ASIZE, P_A16)) {
                 *bytes = 0x67;
-                out(offset, segment, bytes,
-                    OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                 offset += 1;
             } else
                 offset += 0;
@@ -1486,8 +1484,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0311:
             if (bits != 32 && !has_prefix(ins, PPS_ASIZE, P_A32)) {
                 *bytes = 0x67;
-                out(offset, segment, bytes,
-                    OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                 offset += 1;
             } else
                 offset += 0;
@@ -1500,11 +1497,16 @@ static void gencode(int32_t segment, int32_t offset, int bits,
             ins->rex = 0;
             break;
 
+	case 0314:
+	case 0315:
+	case 0316:
+	case 0317:
+	    break;
+
         case 0320:
             if (bits != 16) {
                 *bytes = 0x66;
-                out(offset, segment, bytes,
-                    OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                 offset += 1;
             } else
                 offset += 0;
@@ -1513,8 +1515,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0321:
             if (bits == 16) {
                 *bytes = 0x66;
-                out(offset, segment, bytes,
-                    OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                 offset += 1;
             } else
                 offset += 0;
@@ -1530,7 +1531,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 
         case 0330:
             *bytes = *codes++ ^ condval[ins->condition];
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1540,14 +1541,14 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 	case 0332:
         case 0333:
             *bytes = c - 0332 + 0xF2;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
         case 0334:
             if (ins->rex & REX_R) {
                 *bytes = 0xF0;
-                out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                 offset += 1;
             }
             ins->rex &= ~(REX_L|REX_R);
@@ -1557,15 +1558,13 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 	    break;
 
         case 0340:
-        case 0341:
-        case 0342:
             if (ins->oprs[0].segment != NO_SEG)
                 errfunc(ERR_PANIC, "non-constant BSS size in pass two");
             else {
-                int32_t size = ins->oprs[0].offset << (c & 3);
+                int64_t size = ins->oprs[0].offset;
                 if (size > 0)
                     out(offset, segment, NULL,
-                        OUT_RESERVE + size, NO_SEG, NO_SEG);
+                        OUT_RESERVE, size, NO_SEG, NO_SEG);
                 offset += size;
             }
             break;
@@ -1577,7 +1576,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
         case 0366:
 	case 0367:
             *bytes = c - 0366 + 0x66;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1588,7 +1587,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 
         case 0373:
             *bytes = bits == 16 ? 3 : 5;
-            out(offset, segment, bytes, OUT_RAWDATA + 1, NO_SEG, NO_SEG);
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset += 1;
             break;
 
@@ -1616,6 +1615,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                     errfunc(ERR_NONFATAL, "invalid effective address");
                 }
 
+
                 p = bytes;
                 *p++ = ea_data.modrm;
                 if (ea_data.sib_present)
@@ -1631,8 +1631,7 @@ static void gencode(int32_t segment, int32_t offset, int bits,
 		}
 
                 s = p - bytes;
-                out(offset, segment, bytes, OUT_RAWDATA + s,
-                    NO_SEG, NO_SEG);
+                out(offset, segment, bytes, OUT_RAWDATA, s, NO_SEG, NO_SEG);
 
                 switch (ea_data.bytes) {
                 case 0:
@@ -1640,12 +1639,12 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                 case 1:
                     if (ins->oprs[(c >> 3) & 7].segment != NO_SEG) {
                         data = ins->oprs[(c >> 3) & 7].offset;
-                        out(offset, segment, &data, OUT_ADDRESS + 1,
+                        out(offset, segment, &data, OUT_ADDRESS, 1,
                             ins->oprs[(c >> 3) & 7].segment,
                             ins->oprs[(c >> 3) & 7].wrt);
                     } else {
                         *bytes = ins->oprs[(c >> 3) & 7].offset;
-                        out(offset, segment, bytes, OUT_RAWDATA + 1,
+                        out(offset, segment, bytes, OUT_RAWDATA, 1,
                             NO_SEG, NO_SEG);
                     }
                     s++;
@@ -1654,9 +1653,10 @@ static void gencode(int32_t segment, int32_t offset, int bits,
                 case 2:
                 case 4:
                     data = ins->oprs[(c >> 3) & 7].offset;
+		    warn_overflow(ea_data.bytes, data);
                     out(offset, segment, &data,
-                        (ea_data.rip ?  OUT_REL4ADR : OUT_ADDRESS)
-			+ ea_data.bytes,
+                        ea_data.rip ?  OUT_REL4ADR : OUT_ADDRESS,
+			ea_data.bytes,
                         ins->oprs[(c >> 3) & 7].segment,
                         ins->oprs[(c >> 3) & 7].wrt);
                     s += ea_data.bytes;
