@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version 27/iii/95 by Simon Tatham
@@ -171,7 +171,11 @@ insn *parse_line(int pass, char *buffer, insn * result,
     int critical;
     struct eval_hints hints;
     int j;
+    bool first;
+    bool insn_is_label = false;
 
+restart_parse:
+    first = true;
     result->forw_ref = false;
     error = errfunc;
 
@@ -195,7 +199,9 @@ insn *parse_line(int pass, char *buffer, insn * result,
         return result;
     }
 
-    if (i == TOKEN_ID) {        /* there's a label here */
+    if (i == TOKEN_ID || (insn_is_label && i == TOKEN_INSN)) {
+        /* there's a label here */
+	first = false;
         result->label = tokval.t_charptr;
         i = stdscan(NULL, &tokval);
         if (i == ':') {         /* skip over the optional colon */
@@ -229,6 +235,8 @@ insn *parse_line(int pass, char *buffer, insn * result,
     while (i == TOKEN_PREFIX ||
            (i == TOKEN_REG && !(REG_SREG & ~reg_flags[tokval.t_integer])))
     {
+	first = false;
+
         /*
          * Handle special case: the TIMES prefix.
          */
@@ -259,7 +267,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
 	    int slot = prefix_slot(tokval.t_integer);
 	    if (result->prefixes[slot]) {
                if (result->prefixes[slot] == tokval.t_integer)
- 		    error(ERR_WARNING,
+		    error(ERR_WARNING,
 		      "instruction has redundant prefixes");
                else
 		    error(ERR_NONFATAL,
@@ -335,6 +343,11 @@ insn *parse_line(int pass, char *buffer, insn * result,
             i = stdscan(NULL, &tokval);
             if (i == 0)
                 break;
+	    else if (first && i == ':') {
+		insn_is_label = true;
+		goto restart_parse;
+	    }
+	    first = false;
             fixptr = tail;
             eop = *tail = nasm_malloc(sizeof(extop));
             tail = &eop->next;
@@ -507,6 +520,11 @@ insn *parse_line(int pass, char *buffer, insn * result,
         i = stdscan(NULL, &tokval);
         if (i == 0)
             break;              /* end of operands: get out of here */
+	else if (first && i == ':') {
+	    insn_is_label = true;
+	    goto restart_parse;
+	}
+	first = false;
         result->oprs[operand].type = 0; /* so far, no override */
         while (i == TOKEN_SPECIAL) {    /* size specifiers */
             switch ((int)tokval.t_integer) {
@@ -784,6 +802,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
                     }
                 }
             } else {            /* it's a register */
+		unsigned int rs;
 
                 if (value->type >= EXPR_SIMPLE || value->value != 1) {
                     error(ERR_NONFATAL, "invalid operand type");
@@ -807,16 +826,16 @@ insn *parse_line(int pass, char *buffer, insn * result,
                      * we want to produce a warning iff the specified size
                      * is different from the register size
                      */
-                    i = result->oprs[operand].type & SIZE_MASK;
+                    rs = result->oprs[operand].type & SIZE_MASK;
                 } else
-                    i = 0;
+                    rs = 0;
 
                 result->oprs[operand].type &= TO;
                 result->oprs[operand].type |= REGISTER;
                 result->oprs[operand].type |= reg_flags[value->type];
                 result->oprs[operand].basereg = value->type;
 
-                if (i && (result->oprs[operand].type & SIZE_MASK) != i)
+                if (rs && (result->oprs[operand].type & SIZE_MASK) != rs)
                     error(ERR_WARNING | ERR_PASS1,
                           "register size specification ignored");
             }
